// Copyright (C) 2008-2013 Gael Guennebaud <gael.guennebaud@inria.fr>

/**
 * In this file we declare a basic mesh data structure as an indexed-face-set,
 * build its discrete Laplacian operator,
 * and performs a (bi-)harmonic interpolation of the color components
 */

#include <Eigen/Dense>
#include <Eigen/SparseCore>
#include <Eigen/SparseExtra>

#include <iostream>
#include <limits>
#include <ctime>
#include <vector>
#include <string>
#include <Eigen/Sparse>

/** \class Mesh
  * A class to represent a 3D triangular mesh
  */
class Mesh
{
public:

    /** Default constructor loading a triangular mesh from the file \a filename */
    Mesh(const std::string& filename) {/*...*/}

    /** Destructor */
    virtual ~Mesh() {}
    
    void fillWhiteVertices();
    void computeLaplacianMatrix();

protected:
  
    Eigen::VectorXf MixedVoronoiArea();
    
    typedef Eigen::Vector3f Vector3;

    /** Represents a vertex of the mesh */
    struct Vertex
    {
      Vertex()
        : position(Eigen::Vector3f::Zero()), normal(Eigen::Vector3f::Zero()), color(Eigen::Array3f::Ones()), texcoord(Eigen::Vector2f::Zero())
      {}
      Vertex(const Eigen::Vector3f& pos)
        : position(pos), normal(Eigen::Vector3f::Zero()), color(Eigen::Array3f::Ones()), texcoord(Eigen::Vector2f::Zero())
      {}
      Eigen::Vector3f position;
      Eigen::Vector3f normal;
      Eigen::Array3f  color;
      Eigen::Vector2f texcoord;
    };

    /** Represent a triangular face via its 3 vertex indices. */
    typedef Eigen::Vector3i FaceIndex;

    /** Represents a sequential list of vertices */
    typedef std::vector<Vertex> VertexArray;

    /** Represents a sequential list of triangles */
    typedef std::vector<FaceIndex> FaceIndexArray;

    /** The list of vertices */
    VertexArray mVertices;
    /** The list of face indices */
    FaceIndexArray mFaces;
    
    /** The Laplacian Matrix associated to the mesh */
    Eigen::SparseMatrix<float> mLaplacianMatrix;
};

using namespace Eigen;

// represents a matrix coefficient as a triplet (i,j,value)
typedef Triplet<float> triplet;

/** \returns the mixed voronoi area associated to each vertex */
VectorXf Mesh::MixedVoronoiArea()
{
    
  VectorXf areas(mVertices.size());
  areas.setZero();

  for(int i=0; i< mFaces.size(); i++)
  {
    Vector3f e0 = mVertices[mFaces[i](1)].position - mVertices[mFaces[i](0)].position;
    Vector3f e1 = mVertices[mFaces[i](2)].position - mVertices[mFaces[i](0)].position;

    float triangleArea = 0.5*e0.cross(e1).norm();
    
    // squared edge lengths
    Array3f el2( e0.squaredNorm(), e1.squaredNorm(), (e1-e0).squaredNorm() );
    
    // find the circumcenter restricted to the triangle interior
    Array3f oBar = (el2 * (el2.sum() - 2*el2)).max(0);
    oBar /= oBar.sum();

    // accumulate area distribution
    for(int k=0; k<3; ++k)
      areas[mFaces[i](k)] += triangleArea * 0.5 * (oBar((k+1)%3) + oBar((k+2)%3));
  }
  
  return areas;
}


/** \returns the Laplacian matrix associated to the mesh */
void Mesh::computeLaplacianMatrix()
{
  // Compute areas
  VectorXf areas = MixedVoronoiArea();

  // Find coefficients and create Laplacian Sparse Matrix
  std::vector<triplet> coefficients;

  for(int i=0; i < mFaces.size(); ++i)
  {
    int idxA = mFaces[i](0);
    int idxB = mFaces[i](1);
    int idxC = mFaces[i](2);
    
    Vector3f e0 = mVertices[idxB].position - mVertices[idxA].position;
    Vector3f e1 = mVertices[idxC].position - mVertices[idxA].position;
    float triangleArea = 0.5*e0.cross(e1).norm();

    float coeffAB = 2*e1.dot(e1-e0) / (triangleArea * (areas[idxA] + areas[idxB]));
    float coeffBC = 2*e0.dot(e1)    / (triangleArea * (areas[idxB] + areas[idxC]));
    float coeffCA = 2*e0.dot(e0-e1) / (triangleArea * (areas[idxC] + areas[idxA]));
    
    coefficients.push_back(triplet(idxA, idxB, coeffAB));
    coefficients.push_back(triplet(idxB, idxA, coeffAB));
    coefficients.push_back(triplet(idxA, idxC, coeffCA));
    coefficients.push_back(triplet(idxC, idxA, coeffCA));
    coefficients.push_back(triplet(idxB, idxC, coeffBC));
    coefficients.push_back(triplet(idxC, idxB, coeffBC));

    coefficients.push_back(triplet(idxA, idxA, -coeffCA-coeffAB));
    coefficients.push_back(triplet(idxB, idxB, -coeffAB-coeffBC));
    coefficients.push_back(triplet(idxC, idxC, -coeffBC-coeffCA));
  }

  mLaplacianMatrix.resize(mVertices.size(), mVertices.size());
  mLaplacianMatrix.setFromTriplets(coefficients.begin(), coefficients.end());
}

void Mesh::fillWhiteVertices()
{
  computeLaplacianMatrix();
  
  // Form the permutation to place Unknowns before Fixed vertices
  typedef PermutationMatrix<Dynamic, Dynamic, int> Permutation;
  Permutation pm(mVertices.size());
  int count_unknowns = 0;
  int count_fixed = 0;
  
  for(int i=0 ; i<mVertices.size(); i++)
    if(mVertices[i].color.isOnes())
      pm.indices()[i] = count_unknowns++;

  for(int i=0 ; i<mVertices.size(); i++)
    if(!mVertices[i].color.isOnes())
      pm.indices()[i] = count_unknowns+count_fixed++;
    
  // Apply the permutation to our problem
  // Declare a view on the vertex colors as a Nx3 matrix
  Map<Matrix<float, Dynamic, 3, RowMajor>, 0, OuterStride<11> > colors(mVertices[0].color.data(), mVertices.size(), 3);
  MatrixX3f f = pm * colors;

  SparseMatrix<float> L;
  L = mLaplacianMatrix.twistedBy(pm);
  /// uncomment to perform a bi-harmonic interpolation.
  //L = L * L;
  
  // Factorize L00
  SimplicialLDLT<SparseMatrix<float> > chol(L.topLeftCorner(count_unknowns,count_unknowns));
  
  // Solve L00 x = - L01 * f
  MatrixX3f x = chol.solve(-  L.topRightCorner(count_unknowns, mVertices.size() - count_unknowns)
                            * f.bottomRows(mVertices.size() - count_unknowns));

  if(chol.info()!=Success)
  {
    std::cout << "Solver failed\n";
    return;
  }

  // put the results in the right order
  Permutation pinv = pm.inverse();
  colors = pinv * x;
}


