// g++-mp-4.7 -O2 -msse2 -DNDEBUG demo_cross_sse.cpp -o demo_cross_sse && ./demo_cross_sse

/**
 * This demo demonstrates the performance of AoS versus SoA for the
 * computation of the cross products between one 3D vector B and a
 * set of N points called A.
 * This demo also includes a comparison to a vectorized cross product
 * using Vector4f to store memory aligned 3D vectors (AoS4).
 */
#include <Eigen/Geometry>
#include <bench/BenchTimer.h>
#include <iostream>
using namespace Eigen;

const int N = 32;

typedef Matrix<float,3,N> AoS;
typedef Matrix<float,4,N> AoS4;
typedef Matrix<float,3,N,RowMajor> SoA;

EIGEN_DONT_INLINE void cross_products(const AoS& A, const Vector3f& B, AoS& C)
{
  for(int i=0; i<N; ++i)
    C.col(i) = A.col(i).cross(B);
}

EIGEN_DONT_INLINE void cross_products(const AoS4& A, const Vector4f& B, AoS4& C)
{
  for(int i=0; i<N; ++i)
    C.col(i) = A.col(i).cross3(B);
}

EIGEN_DONT_INLINE void cross_products(const SoA& A, const Vector3f& B, SoA& C)
{
  C = A.colwise().cross(B);
  
  // compiles to:
  //    C.row(0) = A.row(1) * B(2) - A.row(2) * B(1);
  //    C.row(1) = A.row(2) * B(0) - A.row(0) * B(2);
  //    C.row(2) = A.row(0) * B(1) - A.row(1) * B(0);
}

int main()
{
  int tries = 10;
  int rep = 1000000;
  BenchTimer t;
  {
    // AoS
    AoS A, C;
    Vector3f B;
    A.setRandom();
    B.setRandom();
    BENCH(t, tries, rep, cross_products(A, B, C));
    std::cout << "AoS  time: " << t.best() << "s\n";
  }
  {
    // AoS4
    AoS4 A, C;
    Vector4f B;
    A.setRandom();
    B.setRandom();
    BENCH(t, tries, rep, cross_products(A, B, C));
    std::cout << "AoS4 time: " << t.best() << "s\n";
  }
  {
    // SoA
    SoA A, C;
    Vector3f B;
    A.setRandom();
    B.setRandom();
    BENCH(t, tries, rep, cross_products(A, B, C));
    std::cout << "SoA time: " << t.best() << "s\n";
  }
}

