
//  g++-mp-4.7 -O2 -DNDEBUG demo_rbf.cpp -o demo_rbf -I. -L /opt/local/lib/ -lQtCore -lQtGui
//  ./demo_rbf --h 5 --nsamples 1000

/**
 * This demo shows how to assemble and solve a dense least-square problem.
 * 
 * It starts with a R^2->R reference function called 'ref_function',
 * generates a set of random samples (controlled by --nsamples 1000),
 * adds some noise (controlled by --noise 0.1),
 * generates h*h radial basis functions (RBF) center points (called nodes).
 * Use --h 10 to generate 100 nodes.
 * 
 * Then the problem is assembled, solved, and the results is converted into image files:
 * - ref.png -> the reference function
 * - rec.png -> the reconstructed function 
 */

#include <Eigen/Dense>
#include <vector>
#include <iostream>
#include "args.h"
#include <bench/BenchTimer.h>
#include <QtGui/QImage>

using namespace Eigen;

typedef Matrix<double, 2, Dynamic> Matrix2Xd;

/// The R^2 -> R reference function.
double ref_function(double x, double y) {
  using std::cos;
  using std::sin;
  return (x*x-y*y)*0.6 - x*y*0.3 -0.3*x + 0.2*y + 0.55 + 0.3*x*x*y
          + 0.1*sin(18*x)*cos(18*y) + 0.1*sin(8*x)*cos(8*y);
}

int main(int argc, char* argv[])
{
  // loads arguments
  arguments args(argc, argv);  
  int nb_constraints = args.get_int("nsamples", 100);
  double noise = args.get_float("noise", 0.1);
  int h = args.get_int("h", 5);
  int nb_nodes = h*h;
  
  if(nb_nodes>nb_constraints)
  {
    std::cout << "There must be more constraints than RBF nodes" << std::endl;
    return 0;
  }
  
  // generate the samples:
  Matrix2Xd positions(2,nb_constraints);
  VectorXd values(nb_constraints);
  for(int i=0; i<nb_constraints; ++i)
  {
    Array2d x = (Array2d::Random()+1)*0.5;
    positions.col(i) = x;
    values(i) = ref_function(x(0), x(1)) + internal::random<double>(-noise,noise);
  }
  
  // generate the RBF nodes on a regular lattice
  Matrix2Xd nodes(2,nb_nodes);
  for(int i=0; i<h; ++i)
  {
    nodes.row(0).segment(i*h,h).fill((double(i)+0.5)/h);
    nodes.row(1).segment(i*h,h).setLinSpaced(0.5/h,1-0.5/h);
  }
  
  // Assemble the problem into the matrix A. Each row represent one least-square constraint.
  MatrixXd A(nb_constraints, nb_nodes);
  for(int i=0; i<nb_constraints; ++i)
  {
    A.row(i) = (nodes.colwise()-positions.col(i)).colwise().norm().array().cube();
  }
  
  // Solve A*alpha=values in a least-square sense
  BenchTimer t;
  t.start();
  /// Switch between LS solvers:
#if 1
  /// QR factorization + solving
  HouseholderQR<MatrixXd> solver(A);
  VectorXd alpha = solver.solve(values);
#else
  /// Normal equation + Cholesky + solving
  LLT<MatrixXd> solver(A.transpose() * A);
  VectorXd alpha = solver.solve(A.transpose() * values);
#endif
  t.stop();
  std::cout << "Solving time: " << t.value() << std::endl;
  
  // Generate the reconstructed and reference images
  int m = 1000;
  QImage rec(m,m,QImage::Format_RGB32);
  QImage ref(m,m,QImage::Format_RGB32);
  for(int i=0; i<m; ++i)
    for(int j=0; j<m; ++j)
    {
      Vector2d x = (Array2d(i,j)+0.5)/m;
      // Evaluate the RBFs at the location x
      double rec_v = alpha.dot( (nodes.colwise()-x).colwise().norm().array().cube().matrix() );
      double ref_v = ref_function(x(0), x(1));
      rec.setPixel(i,j,qRgb(rec_v*255,rec_v*255,rec_v*255));
      ref.setPixel(i,j,qRgb(ref_v*255,ref_v*255,ref_v*255));
    }
  rec.save("rec.png");
  ref.save("ref.png");
  
  return 0;
}





