// g++-mp-4.7 -O2 -msse2 -DNDEBUG demo_redux.cpp -o demo_redux && ./demo_redux
//

/**
 * This demo demonstrates the benefits of instruction-level parallelism (pipelining).
 * 
 * Compare the three versions of sum_by_hand.
 * 
 * Note that the remaining higher performance of Eigen is due to explicit vectorization.
 */

#include <Eigen/Geometry>
#include <bench/BenchTimer.h>
#include <iostream>
using namespace Eigen;

typedef Matrix<float,Dynamic,1> Vector;


EIGEN_DONT_INLINE float sum_eigen(const Vector& v)
{
  return v.sum();
}

EIGEN_DONT_INLINE float sum_by_hand(const Vector& v)
{
  float ret = 0;
  /// Default version
  for(int i=0; i<v.size(); i++)
    ret = ret + v[i];
  /// Loop peeling (partial unrolling)
//   for(int i=0; i<v.size(); i+=4)
//     ret = ret + v[i] + v[i+1] + v[i+2] + v[i+3];
  /// This version reduce instructions dependencies:
//   for(int i=0; i<v.size(); i+=4)
//     ret = ret + (v[i] + v[i+1]) + (v[i+2] + v[i+3]);
  return ret;
}

int main()
{
  int tries = 10;
  int rep = 1000;
  int n = 10000;
  
  Vector v(n);
  v.setRandom();
    
  BenchTimer t;
  {
    BENCH(t, tries, rep, v(0)=sum_by_hand(v));
    std::cout << "by hand time: " << t.best() << "s  ;  " << 1e-9*(double(rep)*n)/t.best()<< " GFlops\n";
  }
  {
    BENCH(t, tries, rep, v(0)=sum_eigen(v));
    std::cout << "eigen time  : " << t.best() << "s  ;  " << 1e-9*(double(rep)*n)/t.best()<< " GFlops\n";
  }
  return 0;
}

