#include <iostream>

//---------------------------------
// --- Functors for assignments ---
//---------------------------------
template<typename Scalar> struct assign_default { void assign(Scalar* a, const Scalar& b) const { *a  = b; } std::string name() const { return " = "; } };
template<typename Scalar> struct assign_add     { void assign(Scalar* a, const Scalar& b) const { *a += b; } std::string name() const { return " += "; } };

//----------------------------------------
// --- The expression template classes ---
//----------------------------------------

template<typename A, typename B> struct Sum;
template<typename A, typename B> struct Product;
template<typename A, typename B, typename Op> struct Assignment;

/// The base class (like DenseBase/MatrixBase)
template<typename Derived> struct Base
{
  const Derived& derived() const { return *static_cast<const Derived*>(this); }
  Derived& derived() { return *static_cast<Derived*>(this); }
  
  template<typename Other> Sum<Derived,Other> operator+(const Base<Other>& other) const
  { return Sum<Derived,Other>(derived(), other.derived()); }
  
  template<typename Other> Product<Derived,Other> operator*(const Base<Other>& other) const
  { return Product<Derived,Other>(derived(), other.derived()); }
  
  template<typename Other> Derived& operator=(const Base<Other>& other)
  {
    Assignment<Derived,Other,assign_default<int> >::run(derived(), other.derived(), assign_default<int>());
  }
  
  template<typename Other> Derived& operator+=(const Base<Other>& other)
  {
    Assignment<Derived,Other,assign_add<int> >::run(derived(), other.derived(), assign_add<int>());
  }
};

/// A 2x2 Matrix class with storage
struct Matrix : Base<Matrix>
{
  typedef Base<Matrix> Base;
  using Base::operator=;
  Matrix(const std::string& name) : m_name(name) { data = new int[4]; std::cout << "- Create matrix " << m_name << "\n"; }
  ~Matrix() { delete[] data; std::cout << "- Delete matrix " << m_name << "\n"; }
  Matrix(const Matrix& m) { std::cerr << "COPY\n"; }
  
  int coeff(int i) const { return data[i]; }
  int& coeffRef(int i) { return data[i]; }
  
  int size() const { return 4; }
  void setZero() {
    std::cout << "- Set " << m_name << " to zero\n";
    for(int i=0;i<4;++i)
      data[i] = 0;
  }
  
  std::string name() const { return m_name; }
  
  typedef const Matrix& Nested;
  int* data;
  std::string m_name;
};

/// Expression of A + B
template<typename A, typename B> struct Sum : Base<Sum<A,B> >
{
  typedef const Sum Nested;
  Sum(const A& a, const B& b) : m_a(a), m_b(b) {}
  
  std::string name() const { return std::string("(") + m_a.name() + " + " + m_b.name() + ")"; }
  
  typename A::Nested m_a;
  typename B::Nested m_b;
};

/// Expression of A * B
template<typename A, typename B> struct Product : Base<Product<A,B> >
{
  typedef const Product Nested;
  Product(const A& a, const B& b) : m_a(a), m_b(b) {}
  typename A::Nested m_a;
  typename B::Nested m_b;
  
  std::string name() const { return std::string("(") + m_a.name() + " * " + m_b.name() + ")"; }
};

//-------------------
// --- Evaluators ---
//-------------------

/// Default evaluator assuming the expression is it's own evaluator
template<typename T> struct evaluator
{
  typedef evaluator type;
  evaluator(const T& x) : xpr(x) {}
  int coeff(int i) const { return xpr.coeff(i); }
  int& coeffRef(int i) { return const_cast<T&>(xpr).coeffRef(i); }
  const T& xpr;
  
  std::string name() const { return xpr.name(); }
};

/// evaluator of A+B
template<typename A, typename B> struct evaluator<Sum<A,B> >
{
  typedef evaluator type;
  evaluator(const Sum<A,B>& plus) : m_lhs(plus.m_a), m_rhs(plus.m_b) {}
  int coeff(int i) const { return m_lhs.coeff(i) + m_rhs.coeff(i); }
  typename evaluator<A>::type m_lhs;
  typename evaluator<B>::type m_rhs;
  
  std::string name() const { return std::string("(") + m_lhs.name() + " + " + m_rhs.name() + ")"; }
};

/// low level product implementation (c += a*b) - like blas's gemm
void product_impl_low_level(int* c, const int* a, const int* b)
{
  for(int i=0;i<2;++i)
    for(int j=0;j<2;++j)
      for(int k=0;k<2;++k)        
        c[i+j*2] += a[i+k*2] * b[k+j*2];
}

/// special product evaluator (in Eigen it is currently equivalent to blas_traits)
template<typename T> struct product_evaluator
{
  product_evaluator(const T& x)
    : tmp("product_arg_temporary")
  {
    tmp = x;
  }
  const int* data() const { return tmp.data; }
  std::string name() const { return tmp.m_name; }
  Matrix tmp;
};

template<> struct product_evaluator<Matrix>
{
  product_evaluator(const Matrix& x)
    : m_mat(x)
  {}
  const int* data() const { return m_mat.data; }
  std::string name() const { return m_mat.m_name; }
  const Matrix& m_mat;
};

/// performs dest += a * b;
/// A general implementation must be available outisde the evaluator because it is needed by both
/// evaluator<Product> and Assignment<..., Product<...> > for in-place product evaluation
template<typename Dest, typename A, typename B>
void product_impl(Dest& dest, const Product<A,B>& prod)
{
  product_evaluator<A> lhs(prod.m_a);
  product_evaluator<B> rhs(prod.m_b);
  std::cout << "- performs: " << dest.m_name << " += " << lhs.name() << " * " << rhs.name() + " in place\n";
  product_impl_low_level(dest.data, lhs.data(), rhs.data());
};

/// Matrix product evaluator, products are complex and must be evaluated into a temporary
/// unless the evaluator is by-passed by specializations of Assignment<>
template<typename A, typename B>
struct evaluator<Product<A,B> > : evaluator<Matrix>
{
  typedef evaluator<Matrix> Base;
  typedef evaluator type;
  evaluator(const Product<A,B>& prod) : Base(res), res("product_res")
  {
    res.setZero();
    product_impl(res, prod);
  }
  Matrix res;
};


//------------------------
// --- Tree Optimizers ---
//------------------------

/// Default tree optimizer that copies the expression by value
template<typename Xpr> struct TreeOpt
{
  typedef Xpr NewXpr;
  static NewXpr make(const Xpr& xpr) { return xpr; }
};

/// Tree optimizer for Matrix that copies by reference
template<> struct TreeOpt<Matrix>
{
  typedef Matrix NewXpr;
  static const Matrix& make(const Matrix& xpr) { return xpr; }
};

// Needed to forward the optimizer to the children
template<typename A, typename B> struct TreeOpt<Sum<A,B> >
{
  typedef Sum<A,B> Xpr;
  typedef typename TreeOpt<A>::NewXpr NewA;
  typedef typename TreeOpt<B>::NewXpr NewB;
  typedef Sum<NewA,NewB> NewXpr;
  static NewXpr make(const Xpr& xpr) { return TreeOpt<A>::make(xpr.m_a) + TreeOpt<B>::make(xpr.m_b); }
};

// catch A * B + Y and builds Y' + A' * B'
template<typename A, typename B, typename Y>
struct TreeOpt<Sum<Product<A,B>,Y> >
{
  typedef Sum<Product<A,B>, Y> Xpr;
  typedef typename TreeOpt<Y>::NewXpr NewY;
  typedef Sum<NewY,Product<A,B> > NewXpr;
  static NewXpr make(const Xpr& xpr) { return TreeOpt<Y>::make(xpr.m_b) + xpr.m_a; }
};

// catch X + A * B + Y and builds (X' + Y') + (A' * B')
template<typename A, typename B, typename X, typename Y>
struct TreeOpt<Sum<Sum<X,Product<A,B> >,Y> >
{
  typedef Sum<Sum<X,Product<A,B> >,Y> Xpr;
  typedef typename TreeOpt<X>::NewXpr NewX;
  typedef typename TreeOpt<Y>::NewXpr NewY;
  typedef Sum<Sum<NewX,NewY>,Product<A,B> > NewXpr;
  static NewXpr make(const Xpr& xpr) { return (TreeOpt<X>::make(xpr.m_a.m_a) + TreeOpt<Y>::make(xpr.m_b)) + xpr.m_a.m_b; }
};

//-------------------
// --- Assignment ---
//-------------------

template<typename A, typename B, typename Op>
struct Assignment
{
  static void run(A& a, const B& b, const Op& op)
  {
    typename evaluator<A>::type ea(a);
    typename evaluator<B>::type eb(b);
    std::cout << "- default assignment: " << a.name() << op.name() << b.name()<< "\n";
    for(int i=0; i<a.size(); ++i)
      op.assign(&ea.coeffRef(i), eb.coeff(i));
  }
};


// specialization for a += b * c (assuming no aliasing)
template<typename A, typename ProdLhs, typename ProdRhs>
struct Assignment<A,Product<ProdLhs,ProdRhs>,assign_add<int> >
{
  static void run(A& a, const Product<ProdLhs,ProdRhs>& b, const assign_add<int>& )
  {
    product_impl(a, b);
  }
};

// specialization for a = (d) + b * c (assuming no aliasing)
template<typename A, typename D, typename ProdLhs, typename ProdRhs, typename Op>
struct Assignment<A,Sum<D,Product<ProdLhs,ProdRhs> >, Op >
{
  static void run(A& a, const Sum<D,Product<ProdLhs,ProdRhs> >& b, const Op& op)
  {
    Assignment<A,D,Op>::run(a,b.m_a,op);
    Assignment<A,Product<ProdLhs,ProdRhs>, assign_add<int> >::run(a, b.m_b, assign_add<int>() );
  }
};

int main()
{
  Matrix a("a"), b("b"), c("c"), d("d");
  for (int i=0;i<4;++i)
  {
    a.data[i] = i+10;
    b.data[i] = i+20;
    c.data[i] = i+30;
    d.data[i] = i+40;
  }

  auto xpr = a + c*d  + d + c;
  typedef __typeof(xpr) Xpr;
  
  std::cout << "init version:";
  std::cout << " " << xpr.name() << "\n";
  
  
  auto xpr1 = TreeOpt<Xpr>::make(xpr);
  typedef __typeof(xpr1) Xpr1;
  std::cout << "\n optimized version 1:";
  std::cout << " " << xpr1.name() << "\n";
  
  auto xpr2 = TreeOpt<Xpr1>::make(xpr1);
  typedef __typeof(xpr2) Xpr2;
  std::cout << "\n optimized version 2:";
  std::cout << " " << xpr2.name() << "\n";
  
  Matrix r("r");
  
  std::cout << "\nEval r  = xpr;\n";
  r  = xpr;
  
  std::cout << "\nEval r  = xpr2;\n";
  r  = xpr2;
  
//   std::cout << "\nEval r  = a + b*c;\n";
//   r  = a + b*c;
  
  return 0;
}
